/*
 * Copyright (c) 2023–2024 DTB <trinity@trinity.moe>
 * SPDX-License-Identifier: AGPL-3.0-or-later
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Affero General Public License as published by the Free
 * Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see https://www.gnu.org/licenses/.
 */

#include <signal.h> /* sigaction(2), signal(2), struct sigaction, SIGINT */
#include <stdio.h> /* fclose(3), fdopen(3), fprintf(3), getc(3), perror(3),
                    * putc(3), stderr, stdin, stdout, EOF, NULL */
#include <stdlib.h> /* exit(3), size_t, EXIT_FAILURE */
#include <string.h> /* strerror(3) */
#if !defined EX_OK || !defined EX_OSERR || !defined EX_USAGE
#	include <sysexits.h>
#endif
#include <termios.h> /* tcgetattr(3), tcsetattr(3), struct termios, ECHO */
#include <unistd.h> /* getopt(3), isatty(3), STDIN_FILENO */

static int oserr(char *s){ perror(s); return EX_OSERR; }

static char *program_name = "peek";

static int usage(char *s){
	fprintf(stderr, "Usage: %s (-1enot) (-p [program [arguments...]])\n",
		s == NULL ? program_name : s);
	return EX_USAGE;
}

/* Restores terminal echo; otherwise when a user ^Cs the terminal would
 * continue to not display typed text. If sig isn't zero, this will terminate
 * the program. */
void restore_echo(int sig){
	static struct termios t;

	tcgetattr(STDIN_FILENO, &t);
	t.c_lflag |= ECHO;
	tcsetattr(STDIN_FILENO, TCSAFLUSH, &t);

	if(sig != 0)
		exit(EXIT_FAILURE);
	else
		return;
}

int main(int argc, char *argv[]){
	int eof;
	char include_eof;
	FILE *outputs[] = {
		NULL /* stdout */,
		NULL /* stderr */
	};

	if(argc < 1)
		return usage(argv[0]);

	eof = EOF;
	include_eof = 0;
	{ /* options parsing */
		int c;

		while((c = getopt(argc, argv, "1enot")) != -1)
			switch(c){
			case '1':	eof = '\n';          break;
			case 'n':	include_eof = 1;     break;
			case 'o':	outputs[0] = stdout; break;
			case 'e':	outputs[1] = stderr; break;
			case 't':
				if(isatty(STDIN_FILENO) != 1){
					fprintf(stderr, "%s: Must be run in a terminal"
						" (option -t specified)\n", argv[0]);
					return EX_USAGE;
				}
			default:	return usage(argv[0]);
			}

		if(argc > optind)
			return usage(argv[0]);
	}

	{ /* install signal handler */
		/* There isn't a difference in functionality between the signal(2) and
		 * sigaction(2) methods. sigaction(2) is vastly preferred for
		 * portability but some configurations can only use signal(2). */
		/* Errors aren't terminating because the worst that happens is some
		 * terminal phooeyness if things go awry. */
#if defined _POSIX_C_SOURCE
		struct sigaction act = { 0 };

		act.sa_handler = restore_echo;
		if(sigaction(SIGINT, &act, NULL) != 0)
			perror(argv[0]);
#else
		if(signal(SIGINT, restore_echo) == SIG_ERR)
			perror(argv[0]);
#endif
	}
	
	{ /* banish terminal echo */
		struct termios t;

		tcgetattr(STDIN_FILENO, &t);
		t.c_lflag ^= ECHO;
		tcsetattr(STDIN_FILENO, TCSAFLUSH, &t);
	}

	{ /* actual input loop */
		int c;
		size_t i;

		do{	if((c = getc(stdin)) != eof || include_eof)
				for(i = 0; i < (sizeof outputs)/(sizeof *outputs); ++i)
					if(outputs[i] != NULL && putc(c, outputs[i]) == EOF){
						perror(argv[0]);
						outputs[i] = NULL;
					}
		}while(c != eof);
	}

	restore_echo(0);

	return EX_OK;
}
